/*
 *==============================================================================
 * File: daa.c
 *
 * VERSION:
 *
 * Description:
 *     these dynamic array allocators - daa() and daam() - are designed
 *     to be efficient and flexible.  daa() takes as argument a pointer
 *     to the space that the caller allocates.  the size of this space is
 *     determined by a previous call to das().  the usual sequence would
 *     be das()/malloc()/daa().  daam() takes as argument a pointer to
 *     pointer that daam() returns the pointer to the malloc()'ed space in.
 *     they can allocate arrays of up to MAX_DIM dimensions of any type that
 *     will return a size with the sizeof() C operator.  they are also very
 *     efficient from the point of view of array access.  for daam() the
 *     single malloc() ensures locality of reference that eliminates excess
 *     paging encountered with dynamic array allocation schemes that use
 *     multiple malloc()'s.  arrays of structure, enum or union type can
 *     be allocated.  a corresponding free of the allocated area must
 *     normally be done (unless you want to allocate to program termination).
 *     for daam() the sixth parameter returns the free pointer; daa() does
 *     not require a free pointer since the caller allocates the space.
 *     do not free on the function return pointer since array storage begins
 *     with the data and is followed by pointers to pointers to ..., and the
 *     first element(lowest subscript of each dimension), does not point to
 *     the beginning of allocated space.  the array is initialized to the
 *     value pointed to by the last parameter, or not initialized if NULL.
 *     the last parameter init pointer should point to something with a size
 *     the same as the size of the type in the sizeof() first argument.
 *     arrays may be allocated to have one or more dimensions with non-zero
 *     integer(+ or -) start subscripts.  thus, arrays may be one based or
 *     zero based or minus one based or any based for that matter.
 *
 * Examples:
 *     see the test suite at the end of this file for test examples.
 *
 * Author:
 *    Richard Hogaboom
 *
 *==============================================================================
 */

/*
 *=================================================================*
 *                                                                 *
 *   (c) Copyright, 2001 Massachusetts Institute of Technology.    *
 *       This material may be reproduced by or for the             *
 *       U.S. Government pursuant to the copyright license         *
 *       under the clause at 252.227-7013 (June, 1995).            *
 *                                                                 *
 *=================================================================*
 */

/*
 * Notes:
 *     1. the dimensional limit of the allocated arrays is set by MAX_DIM;
 *        increase or decrease this to adjust the maximum array dimension.
 *     2. this code has been tested(all suite tests) with the SunPro(cc) and
 *        Gnu(gcc) compilers with a Sun as the target system.  the commands are:
 *
 *           cc -Xc -g -DDAA_TEST -o daa daa.c -lm
 *           gcc -g -ansi -pedantic -Wall -DDAA_TEST -o daa daa.c -lm
 *
 *     3. alignment of the data area is the alignment of the first argument
 *        sizeof(), while the alignment of the pointer area is sizeof(char *).
 *        the data comes first, and the assumption here is that the malloc()
 *        (for daam()) and whatever allocation routine is used between das()
 *        and daa() will align at the most stringent boundary, thus
 *        accommodating the data area alignment.  the pointer area comes
 *        second and may, depending on the total size of the data area need
 *        to be aligned on a sizeof(char *) boundary.  the beginning of the
 *        pointer area is tested for alignment, and its alignment adjusted if
 *        necessary.
 *     4. the use of the mtmalloc library for memory allocation was motivated
 *        by problems encountered in a multithreaded environment with the
 *        standard Sun malloc library that caused difficult to trace crashes.
 *        the thread hot mtmalloc library seemed to cure this problem.
 *        experimentation on other machine architectures and operating systems
 *        is desirable and necessary.
 *     5. for an excellent reference on this type of array access see Numerical
 *        Recipes in C, Press, Flannery, Teukolsky, and Vettering, Cambridge
 *        University Press, 1992, pg. 20.
 */

/*
 *------------------------------------------------------------------------------
 * include files
 *------------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <mtmalloc.h>

#include "daa.h"

/*
 *------------------------------------------------------------------------------
 * constants
 *------------------------------------------------------------------------------
 */

/* global constants */
#define MAX_DIM 10 /* maximum number of array dimensions */

/* error msg indices */
#define ERRS_INV_DIMS      0
#define ERRS_INV_REQ_SIZE  1
#define ERRS_INV_DIM       2
#define ERRS_FAIL_ALLOC    3

/*
 *------------------------------------------------------------------------------
 * data types
 *------------------------------------------------------------------------------
 */

/*
 *------------------------------------------------------------------------------
 * global variables
 *------------------------------------------------------------------------------
 */

/* error msgs */
char *daa_errs[] =
{
    "daa: invalid number of dimensions - must be > 0 and <= MAX_DIM.",
    "daa: invalid request size - must be > 0.",
    "daa: invalid dimension - must be > 0.",
    "daa: failure to allocate memory.",
};

/*
 *------------------------------------------------------------------------------
 * local function prototypes
 *------------------------------------------------------------------------------
 */

/*
 *------------------------------------------------------------------------------
 * functions
 *------------------------------------------------------------------------------
 */



/*
 *------------------------------------------------------------------------------
 * doff:
 *     data offset routine that calculates the offset in basic data
 *     units of the location of the data item specified by the
 *     current settings in dim_ind[].  this multiplied by the size of
 *     a basic data unit plus the offset of the start of the area
 *     gives the address of the data item.  a basic data unit may be
 *     either a pointer(sizeof(char *) bytes long) or array data
 *     item(data_size bytes long).
 *
 * Arguments:
 *     int level
 *        the array level to initialize pointers for.  the first
 *        dimension is level 0, the second level 1, etc., always
 *        starts at zero.
 *
 *     int *dim_ind
 *        dimension index array of current call, starts zeroed.
 *
 *     int dim_prod
 *        product of dimensions from zero to level.
 *
 *     int *dp
 *        array of dimension products.
 *
 * Returns:
 *     int offset in basic data units from the start of the data area(base_ptr)
 *     to data item specified by dim_ind[].  byte offset is given by:
 *     doff(level, dim_ind, dim_prod, dp) * data_size.
 *------------------------------------------------------------------------------
 */

static int
doff(int level, int *dim_ind, int dim_prod, int *dp)
{
    if (level == 0)
    {
        return 0;
    }
    else
    {
        return doff(level-1, dim_ind, dim_prod, dp) +
            (dim_prod/dp[level-1]) * dim_ind[level-1];
    }
}



/*
 *------------------------------------------------------------------------------
 * off:
 *     offset routine that calculates the offset in pointer units(char *)
 *     of the start of the pointer array for that level.  each level has a
 *     pointer array associated with that level that points to the next level
 *     and is pointed to by the previous level.
 *
 * Arguments:
 *     int level
 *        the array level to initialize pointers for.  the first dimension is
 *        level 0, the second level 1, etc., always starts at zero.
 *
 *     int *dp
 *        array of dimension products.
 *
 * Returns:
 *     int offset in pointer units(char *) of the pointer array for that level.
 *     byte offset is given by off(level, dp) * sizeof(char *).
 *------------------------------------------------------------------------------
 */

static int
off(int level, int *dp)
{
    if (level == 0)
    {
        return 0;
    }
    else
    {
        return dp[level-1] + off(level-1, dp);
    }
}



/*
 *------------------------------------------------------------------------------
 * ptr_init:
 *     this routine uses the space pointed to by the "base_ptr" pointer.
 *     it constructs the several levels of pointers needed to reference
 *     elements of a multidimensional array.
 *
 * Arguments:
 *     int level
 *        the array level to initialize pointers for.  the first dimension
 *        is level 0, the second level 1, etc., always starts at zero.
 *
 *     int *dim_ind
 *        dimension index array of current call, starts zeroed.
 *
 *     int data_size
 *        size of the basic array data object.  this will usually be
 *        obtained from the sizeof() operator.
 *
 *     int num_dim
 *        number of array dimensions.
 *
 *     char *data_ptr
 *        points to base of allocated array.
 *
 *     char *ptr_ptr
 *        points to base of pointers to pointers to ... to data
 *
 *     int *dim
 *        a single dimensional int array of the dimensions of the array to
 *        be allocated.
 *
 *     int *st
 *        a single dimensional int array of integer start subscripts for
 *        each corresponding dimension of the dim array, elements may
 *        be negative.
 *
 *     int *dp
 *        array of dimension products.
 *
 * Returns:
 *     pointer to next level of indirection.
 *------------------------------------------------------------------------------
 */

static char *
ptr_init(int level, int *dim_ind, int data_size, int num_dim,
    char *data_ptr, char *ptr_ptr, int *dim, int *st, int *dp)
{
    char		**ptrs;
    unsigned long	i;



    /*
     * if level is not final level get pointer to array of pointers
     * for that level and recursively call ptr_init() to return pointers to
     * fill the array of pointers.  if level is final level calculate
     * position of data and return pointer to level-1.
     */

    if (level < (num_dim - 1))
    {
        /*
         * ptrs points to the start of a subarray of pointers to the
         * next dimension level of pointers or data
         */

        ptrs = (char **) (ptr_ptr + off(level, dp) * sizeof(char *) +
            ((level == 0)?0:doff(level, dim_ind,
            dp[level], dp) * sizeof(char *)));

        /* fill the array of pointers that ptrs points to */
        for (i=0 ; i<dim[level] ; i++)
        {
            dim_ind[level] = i;

            /*
             * the recursive call to ptr_init() returns a pointer to the
             * next level and is then adjusted by st[] for a
             * non-zero based subscript
             */

            ptrs[i] = ptr_init(level+1, dim_ind, data_size, num_dim, data_ptr,
                ptr_ptr, dim, st, dp) - st[level+1] *
                ((level+1 == num_dim-1)?data_size:sizeof(char *));
        }

        /*
         * adjust zeroth level pointer that is passed back to root
         * of recursive tree call
         */

        if (level == 0)
        {
            ptrs -= st[0];
        }
    }
    else
    {
        /*
         * calculate pointers to data subarrays.  if one dimensional
         * array and non-zero subscripting adjust returned pointer
         */

        ptrs = (char **) (data_ptr + doff(level, dim_ind, dp[level], dp) *
            data_size - ((level == 0)?(st[0] * data_size):0));
    }

    return (char *) ptrs;
}



/*
 *------------------------------------------------------------------------------
 * das:
 *     dynamic array size.  this routine takes four of the same arguments
 *     that daa() and daam() take and calculates the total heap allocation
 *     in bytes required to store the array.  normally used in conjunction
 *     with daa() to do das()/malloc()/daa() sequence.
 *
 * Arguments:
 *     int data_size
 *        size of the basic array data object.  this will usually be
 *        obtained from the sizeof() operator.
 *
 *     int num_dim
 *        number of array dimensions.
 *
 *     int *dim
 *        a single dimensional int array of the dimensions of the array to
 *        be allocated.
 *
 *     int *err_code
 *        index to returned error code string in daa_errs[].
 *
 * Returns:
 *     size in bytes of dynamic array that daa() and daam() will use.
 *------------------------------------------------------------------------------
 */

int
das(int data_size, int num_dim, int *dim, int *err_code)
{
    int i;

    /* product of dimensions from 0 to index, dim[0]*dim[1]...dim[index] */
    int dp[MAX_DIM];



    if (num_dim < 1 || num_dim > MAX_DIM)
    {
        *err_code = ERRS_INV_DIMS;
        return -1;
    }

    if (data_size < 1)
    {
        *err_code = ERRS_INV_REQ_SIZE;
        return -1;
    }

    dp[0] = dim[0];
    for (i=0 ; i<num_dim ; i++)
    {
        if (dim[i] <= 0)
        {
            *err_code = ERRS_INV_DIM;
            return -1;
        }

        if (i > 0)
        {
            dp[i] = dp[i-1] * dim[i];
        }
    }

    return(dp[num_dim-1] * data_size + off(num_dim-1, dp) * sizeof(char *) +
        sizeof(char *));
}



/*
 *------------------------------------------------------------------------------
 * daa:
 *     dynamic array allocation with memory allocated externally.
 *
 * Arguments:
 *     int data_size
 *     size of the basic array data object.  this will usually be
 *     obtained from the sizeof() operator.
 *
 *     int num_dim
 *        number of array dimensions.
 *
 *     int *dim
 *        a single dimensional int array of the dimensions of the array to
 *        be allocated.
 *
 *     int *st
 *        a single dimensional int array of integer start subscripts for
 *        each corresponding dimension of the dim array, elements may
 *        be negative.
 *
 *     int *err_code
 *        index to returned error code string in daa_errs[].
 *
 *     char *base_ptr
 *        pointer to the array space.
 *
 *     char *init_ptr
 *        initialization pointer parameter.
 *
 * Returns:
 *     error free operation returns a pointer to void that points to the start
 *     of the dynamically allocated array area and does not set *err_code.  the
 *     returned pointer will need to be cast to the type of the subsequent array
 *     references.
 *
 *     routine failure returns NULL and any of the following
 *     error codes in *err_code:
 *     ERRS_INV_DIMS - invalid number of dimensions - must be > 0 and <= MAX_DIM.
 *     ERRS_INV_REQ_SIZE - invalid request size - must be > 0.
 *     ERRS_INV_DIM - invalid dimension - must be > 0.
 *     ERRS_FAIL_ALLOC - failure to allocate memory. (daam() only)
 *
 *     failure to index the subscripts in the manner established by the starting
 *     index and the dimensional extent arrays will of course result in run time
 *     errors.  this is not an array allocation error but an array usage error,
 *     similar to any array access on a static C zero based array outside the
 *     normal 0...n-1 bounds.
 *------------------------------------------------------------------------------
 */

void *
daa(int data_size, int num_dim, int *dim, int *st,
    int *err_code, char *base_ptr, char *init_ptr)
{
    int		i, j;

    /* array of current dimension indices */
    int		dim_ind[MAX_DIM];

    char	*p_data, /* pointer to array data */
        	*p;      /* tmp pointer */

    /* product of dimensions from 0 to index, dim[0]*dim[1]...dim[index] */
    int		dp[MAX_DIM];

    /* points to base of pointers to pointers to ... to data */
    char	*ptr_ptr;



    if (num_dim < 1 || num_dim > MAX_DIM)
    {
        *err_code = ERRS_INV_DIMS;
        return NULL;
    }

    if (data_size < 1)
    {
        *err_code = ERRS_INV_REQ_SIZE;
        return NULL;
    }

    /*
     * set dim_ind to all zeros, the multiple invocations of the recursive
     * array allocation routine ptr_init() will pass changed by 1 dim_ind arrays
     * to each invocation of ptr_init().  set dp[] from dim[] input array.
     */

    dp[0] = dim[0];
    for (i=0 ; i<num_dim ; i++)
    {
        dim_ind[i] = 0;
        if (dim[i] <= 0)
        {
            *err_code = ERRS_INV_DIM;
            return NULL;
        }

        if (i > 0)
        {
            dp[i] = dp[i-1] * dim[i];
        }
    }

    /*
     * calculate address of start of pointers.  If not
     * sizeof(char *) aligned make it so.
     */
    ptr_ptr = base_ptr + dp[num_dim-1] * data_size;

    for (i=0 ; i<sizeof(char *) ; i++, ++ptr_ptr)
    {
        if (((int)ptr_ptr)%sizeof(char *) == 0)
        {
            break;
        }
    }

    /* if init_ptr is NULL skip initialization */
    if (init_ptr != NULL)
    {
        /* set p_data to point to the start of the data area */
        p_data = base_ptr;
     
        /* initialize the array */
        for (i=0 ; i<dp[num_dim-1] ; i++)
        {
            p = init_ptr;
            for (j=0 ; j<data_size ; j++)
            {
                *p_data++ = *p++;
            }
        }
    }

    /* do array setup i.e. all the pointer stuff */
    return ptr_init(0, dim_ind, data_size, num_dim, base_ptr,
        ptr_ptr, dim, st, dp);
}



/*
 *------------------------------------------------------------------------------
 * daam:
 *     dynamic array allocation with memory allocated by malloc().
 *
 * Arguments:
 *     int data_size
 *     size of the basic array data object.  this will usually be
 *     obtained from the sizeof() operator.
 *
 *     int num_dim
 *        number of array dimensions.
 *
 *     int *dim
 *        a single dimensional int array of the dimensions of the array to
 *        be allocated.
 *
 *     int *st
 *        a single dimensional int array of integer start subscripts for
 *        each corresponding dimension of the dim array, elements may
 *        be negative.
 *
 *     int *err_code
 *        index to returned error code string in daa_errs[].
 *
 *     char char **free_ptr
 *        pointer to the base pointer of the malloc()'ed array
 *        space for the caller to free.
 *
 *     char *init_ptr
 *        initialization pointer parameter.
 *
 * Returns:
 *     error free operation returns a pointer to void that points to the start
 *     of the dynamically allocated array area and does not set *err_code.  the
 *     returned pointer will need to be cast to the type of the subsequent array
 *     references.
 *
 *     routine failure returns NULL and any of the following
 *     error codes in *err_code:
 *     ERRS_INV_DIMS - invalid number of dimensions - must be > 0 and <= MAX_DIM.
 *     ERRS_INV_REQ_SIZE - invalid request size - must be > 0.
 *     ERRS_INV_DIM - invalid dimension - must be > 0.
 *     ERRS_FAIL_ALLOC - failure to allocate memory. (daam() only)
 *
 *     failure to index the subscripts in the manner established by the starting
 *     index and the dimensional extent arrays will of course result in run time
 *     errors.  this is not an array allocation error but an array usage error,
 *     similar to any array access on a static C zero based array outside the
 *     normal 0...n-1 bounds.
 *------------------------------------------------------------------------------
 */

void *
daam(int data_size, int num_dim, int *dim, int *st,
    int *err_code, char **free_ptr, char *init_ptr)
{
    int		i, j;

    /* array of current dimension indices */
    int dim_ind[MAX_DIM];

    char	*p_data, /* pointer to array data */
        	*p;      /* tmp pointer */

    /* points to base of allocated array */
    char	*base_ptr;

    /* points to base of pointers to pointers to ... to data */
    char	*ptr_ptr;

    /* product of dimensions from 0 to index, dim[0]*dim[1]...dim[index] */
    int		dp[MAX_DIM];



    if (num_dim < 1 || num_dim > MAX_DIM)
    {
        *err_code = ERRS_INV_DIMS;
        return NULL;
    }

    if (data_size < 1)
    {
        *err_code = ERRS_INV_REQ_SIZE;
        return NULL;
    }

    /*
     * set dim_ind to all zeros, the multiple invocations of the recursive
     * array allocation routine ptr_init() will pass changed by 1 dim_ind arrays
     * to each invocation of ptr_init().  set dp[] from dim[] input array.
     */

    dp[0] = dim[0];
    for (i=0 ; i<num_dim ; i++)
    {
        dim_ind[i] = 0;
        if (dim[i] <= 0)
        {
            *err_code = ERRS_INV_DIM;
            return NULL;
        }

        if (i > 0)
        {
            dp[i] = dp[i-1] * dim[i];
        }
    }

    /*
     * allocate enough memory for the data(dp[num_dim-1]*data_size) plus all
     * the array pointers(off(num_dim-1, dp)*sizeof(char *)) plus an extra
     * sizeof(char *) to allow for possible realignment of start of pointer
     * area on pointer alignment boundary.
     */

    if ((base_ptr = (char *) malloc(dp[num_dim-1] * data_size +
           off(num_dim-1, dp) * sizeof(char *) + sizeof(char *))) == NULL)
    {
        *err_code = ERRS_FAIL_ALLOC;
        return NULL;
    }

    /*
     * calculate address of start of pointers.  If not
     * sizeof(char *) aligned make it so.
     */
    ptr_ptr = base_ptr + dp[num_dim-1] * data_size;

    for (i=0 ; i<sizeof(char *) ; i++, ++ptr_ptr)
    {
        if (((int)ptr_ptr)%sizeof(char *) == 0)
        {
            break;
        }
    }

    /* return allocated space pointer that caller should use to free space */
    *free_ptr = base_ptr;

    /* if init_ptr is NULL skip initialization */
    if (init_ptr != NULL)
    {
        /* set p_data to point to the start of the data area */
        p_data = base_ptr;
     
        /* initialize the array */
        for (i=0 ; i<dp[num_dim-1] ; i++)
        {
            p = init_ptr;
            for (j=0 ; j<data_size ; j++)
            {
                *p_data++ = *p++;
            }
        }
    }

    /* do array setup i.e. all the pointer stuff */
    return ptr_init(0, dim_ind, data_size, num_dim, base_ptr,
        ptr_ptr, dim, st, dp);
}



#ifdef DAA_TEST

#include <string.h>

/*
 * test code for daa(), daam(), and das().  all the tests are completely
 * independent, that is, any test may be eliminated or moved elsewhere and
 * run independently.
 */

   void
test1(int d[4], int st[4], double ****array)
{
    int i, j, k, l;

    for (i=st[0] ; i<st[0]+d[0] ; i++)
    {
        for (j=st[1] ; j<st[1]+d[1] ; j++)
        {
            for (k=st[2] ; k<st[2]+d[2] ; k++)
            {
                for (l=st[3] ; l<st[3]+d[3] ; l++)
                {
                    array[i][j][k][l] = i + j + k + l;
                }
            }
        }
    }
}



#define DAA  0
#define DAAV  1

   int
main()
{
    int test;



    /* to test daa() set test to DAA; to test daam() set test to DAAV */
    test = DAA;

    /*
     * TEST 1
     *    4 dimensional double array with subroutine call(daam())
     *    modification of array.
     */
    {
        int i, j, k, l;
        int err_code = 0;
        int asize = 0;
        char *free_ptr;
        int d[4] = {3, 5, 4, 2};
        int st[4] = {-1, -5, 10, 0};
        double ****array;

        fprintf(stderr, "\nTEST 1\n");

        asize = das(sizeof(double), 4, d, &err_code);

        if (test == DAA)
        {
            free_ptr = (char *)malloc(asize);

            array = (double ****) daa(sizeof(double), 4, d, st,
                &err_code, free_ptr, NULL);
        }
        else
        {
            array = (double ****) daam(sizeof(double), 4, d, st,
                &err_code, &free_ptr, NULL);
        }

        if (array == NULL)
        {
            fprintf(stderr, "daa(m): error on dynamic allocation. %s\n"
                , daa_errs[err_code]);
        }
        else
        {
            test1(d, st, array);

            fprintf(stderr, "sizeof(double) = %d\n", sizeof(double));
            fprintf(stderr, "array size = %d\n", asize);
            for (i=st[0] ; i<st[0]+d[0] ; i++)
            {
                for (j=st[1] ; j<st[1]+d[1] ; j++)
                {
                    for (k=st[2] ; k<st[2]+d[2] ; k++)
                    {
                        for (l=st[3] ; l<st[3]+d[3] ; l++)
                        {
                            fprintf(stderr,
                                "array[%2d][%2d][%2d][%2d] = %5.1f\n"
                                , i, j, k, l, array[i][j][k][l]);
                        }
                    }
                }
            }
            fprintf(stderr, "err_code = %d\n", err_code);
            free(free_ptr);
        }
    }

    /*
     * TEST 2
     *    4 dimensional double array with subroutine call(daa())
     *    modification of array.
     */
    {
        int i, j, k, l;
        int err_code = 0;
        int asize = 0;
        char *free_ptr;
        int d[4] = {3, 5, 4, 2};
        int st[4] = {-1, -5, 10, 0};
        double init = 123.;
        double ****array;

        fprintf(stderr, "\nTEST 2\n");

        asize = das(sizeof(double), 4, d, &err_code);

        if (test == DAA)
        {
            free_ptr = (char *)malloc(asize);

            array = (double ****) daa(sizeof(double), 4, d, st,
                &err_code, free_ptr, (char *)&init);
        }
        else
        {
            array = (double ****) daam(sizeof(double), 4, d, st,
                &err_code, &free_ptr, (char *)&init);
        }

        if (array == NULL)
        {
            fprintf(stderr, "daa(m): error on dynamic allocation. %s\n"
                , daa_errs[err_code]);
        }
        else
        {
            test1(d, st, array);

            fprintf(stderr, "sizeof(double) = %d\n", sizeof(double));
            fprintf(stderr, "array size = %d\n", asize);
            for (i=st[0] ; i<st[0]+d[0] ; i++)
            {
                for (j=st[1] ; j<st[1]+d[1] ; j++)
                {
                    for (k=st[2] ; k<st[2]+d[2] ; k++)
                    {
                        for (l=st[3] ; l<st[3]+d[3] ; l++)
                        {
                            fprintf(stderr,
                                "array[%2d][%2d][%2d][%2d] = %5.1f\n"
                                , i, j, k, l, array[i][j][k][l]);
                        }
                    }
                }
            }
            fprintf(stderr, "err_code = %d\n", err_code);
            free(free_ptr);
        }
    }

    /*
     * TEST 3
     *    1 dimensional double array.
     */
    {
        int i;
        int err_code = 0;
        int asize = 0;
        char *free_ptr;
        int d[1] = {10};
        int st[1] = {1};
        double init = 10.10;
        double *array;

        fprintf(stderr, "\nTEST 3\n");

        asize = das(sizeof(double), 1, d, &err_code);

        if (test == DAA)
        {
            free_ptr = (char *)malloc(asize);

            array = (double *) daa(sizeof(double), 1, d, st,
               &err_code, free_ptr, (char *)&init);
        }
        else
        {
            array = (double *) daam(sizeof(double), 1, d, st,
               &err_code, &free_ptr, (char *)&init);
        }

        if (array == NULL)
        {
            fprintf(stderr, "daa(m): error on dynamic allocation. %s\n"
                , daa_errs[err_code]);
        }
        else
        {
            array[10] = 4.5;

            fprintf(stderr, "sizeof(double) = %d\n", sizeof(double));
            fprintf(stderr, "array size = %d\n", asize);
            for (i=st[0] ; i<st[0]+d[0] ; i++)
            {
                fprintf(stderr, "array[%2d] = %e\n", i, array[i]);
            }
            fprintf(stderr, "err_code = %d\n", err_code);
            free(free_ptr);
        }
    }

    /*
     * TEST 4
     *    10 dimensional double array.
     *
     * Note:
     *    this is by far the largest allocation test.  only the first and
     *    last array elements are printed out.  they should be 1.0 and
     *    368640.0 respectively.  the last number is obtained by taking
     *    the product of all the dimensions(d[]).
     */
    {
        int i0, i1, i2, i3, i4, i5, i6, i7, i8, i9;
        int err_code = 0;
        int asize = 0;
        char *free_ptr;
        int d[10] = {3, 3, 2, 4, 5, 4, 4, 4, 4, 4};
        int st[10] = {-4, -3, -2, -1, 0, 1, 2, 3, 4, 5};
        double data;
        double init = 0;
        double **********array;

        fprintf(stderr, "\nTEST 4\n");

        asize = das(sizeof(double), 10, d, &err_code);

        if (test == DAA)
        {
            free_ptr = (char *)malloc(asize);

            array = (double **********) daa(sizeof(double), 10, d, st,
                &err_code, free_ptr, (char *)&init);
        }
        else
        {
            array = (double **********) daam(sizeof(double), 10, d, st,
                &err_code, &free_ptr, (char *)&init);
        }

        if (array == NULL)
        {
            fprintf(stderr, "daa(m): error on dynamic allocation. %s\n"
                ,daa_errs[err_code]);
        }
        else
        {
            data = 0.;
            for (i0=st[0] ; i0<st[0]+d[0] ; i0++)
            {
	       for (i1=st[1] ; i1<st[1]+d[1] ; i1++)
	       {
		  for (i2=st[2] ; i2<st[2]+d[2] ; i2++)
		  {
		     for (i3=st[3] ; i3<st[3]+d[3] ; i3++)
		     {
			for (i4=st[4] ; i4<st[4]+d[4] ; i4++)
			{
			   for (i5=st[5] ; i5<st[5]+d[5] ; i5++)
			   {
			      for (i6=st[6] ; i6<st[6]+d[6] ; i6++)
			      {
				 for (i7=st[7] ; i7<st[7]+d[7] ; i7++)
				 {
				     for (i8=st[8] ; i8<st[8]+d[8] ; i8++)
				     {
					for (i9=st[9] ; i9<st[9]+d[9] ; i9++)
					{
					    array[i0][i1][i2][i3][i4]
                                                 [i5][i6][i7][i8][i9] = ++data;
					}
				     }
				 }
			      }
			   }
			}
		     }
		  }
	       }
            }

            fprintf(stderr, "sizeof(double) = %d\n", sizeof(double));
            fprintf(stderr, "array size = %d\n", asize);
            fprintf(stderr, "array[-4][-3][-2][-1][0][1][2][3][4][5] = %e\n"
                ,array[-4][-3][-2][-1][0][1][2][3][4][5]);
            fprintf(stderr, "array[-2][-1][-1][2][4][4][5][6][7][8] = %e\n"
                , array[-2][-1][-1][2][4][4][5][6][7][8]);
            fprintf(stderr, "err_code = %d\n", err_code);
            free(free_ptr);
        }
    }

    /*
     * TEST 5
     *    1 dimensional array of struct.
     */
    {
        int i;
        int err_code = 0;
        int asize = 0;
        char *free_ptr;
        int d[1] = {10};
        int st[1] = {1};
        struct s {
            double d;
            int l;
        } *array, s_init = {1.25, 5};

        fprintf(stderr, "\nTEST 5\n");

        asize = das(sizeof(struct s), 1, d, &err_code);

        if (test == DAA)
        {
            free_ptr = (char *)malloc(asize);

            array = (struct s *) daa(sizeof(struct s), 1, d, st,
                &err_code, free_ptr, (char *)&s_init);
        }
        else
        {
            array = (struct s *) daam(sizeof(struct s), 1, d, st,
                &err_code, &free_ptr, (char *)&s_init);
        }

        if (array == NULL)
        {
            fprintf(stderr, "daa(m): error on dynamic allocation. %s\n"
                , daa_errs[err_code]);
        }
        else
        {
            array[1].l= 10;
            array[1].d= .5;
            array[10].l = 4;
            array[10].d = 4.5;

            fprintf(stderr, "sizeof(struct s) = %d\n", sizeof(struct s));
            fprintf(stderr, "array size = %d\n", asize);
            for (i=st[0] ; i<st[0]+d[0] ; i++)
            {
                fprintf(stderr, "array[%2d].d = %f\n", i, array[i].d);
                fprintf(stderr, "array[%2d].l = %2d\n", i, array[i].l);
            }
            fprintf(stderr, "err_code = %d\n", err_code);
            free(free_ptr);
        }
    }

    /*
     * TEST 6
     *    2 dimensional array of struct.
     */
    {
        int i, j;
        int err_code = 0;
        int asize = 0;
        char *free_ptr;
        int d[2] = {5, 5};
        int st[2] = {0, 0};
        struct s {
            double d;
            int l;
        } **array, s_init = {1, 5};

        fprintf(stderr, "\nTEST 6\n");

        asize = das(sizeof(struct s), 2, d, &err_code);

        if (test == DAA)
        {
            free_ptr = (char *)malloc(asize);

            array = (struct s **) daa(sizeof(struct s), 2, d, st,
                &err_code, free_ptr, (char *)&s_init);
        }
        else
        {
            array = (struct s **) daam(sizeof(struct s), 2, d, st,
                &err_code, &free_ptr, (char *)&s_init);
        }

        if (array == NULL)
        {
            fprintf(stderr, "daa(m): error on dynamic allocation. %s\n"
                , daa_errs[err_code]);
        }
        else
        {
            array[0][0].l = 10;
            array[0][0].d = .25;
            array[4][4].l = 4;
            array[4][4].d = 4.5;

            fprintf(stderr, "sizeof(struct s) = %d\n", sizeof(struct s));
            fprintf(stderr, "array size = %d\n", asize);
            for (i=st[0] ; i<st[0]+d[0] ; i++)
            {
                for (j=st[1] ; j<st[1]+d[1] ; j++)
                {
                    fprintf(stderr,
                        "array[%2d][%2d].l = %2d array[%2d][%2d].d = %f\n",
                        i, j, array[i][j].l, i, j, array[i][j].d);
                }
            }
            fprintf(stderr, "err_code = %d\n", err_code);
            free(free_ptr);
        }
    }

    /*
     * TEST 7
     *    3 dimensional array of struct.
     */
    {
        int i, j, k;
        int err_code = 0;
        int asize = 0;
        char *free_ptr;
        int d[3] = {2, 5, 6};
        int st[3] = {0, 0, 0};
        struct s {
            char c;
            int l;
            double d;
        } ***array, s_init = {0, 1, 10};

        fprintf(stderr, "\nTEST 7\n");

        asize = das(sizeof(struct s), 3, d, &err_code);

        if (test == DAA)
        {
            free_ptr = (char *)malloc(asize);

            array = (struct s ***) daa(sizeof(struct s), 3, d, st,
                &err_code, free_ptr, (char *)&s_init);
        }
        else
        {
            array = (struct s ***) daam(sizeof(struct s), 3, d, st,
                &err_code, &free_ptr, (char *)&s_init);
        }

        if (array == NULL)
        {
            fprintf(stderr, "daa(m): error on dynamic allocation. %s\n"
                , daa_errs[err_code]);
        }
        else
        {
            array[1][4][5].c = 1;
            array[1][4][5].l = 15;
            array[1][4][5].d = 7.5;

            fprintf(stderr, "sizeof(struct s) = %d\n", sizeof(struct s));
            fprintf(stderr, "array size = %d\n", asize);
            for (i=st[0] ; i<st[0]+d[0] ; i++)
            {
                for (j=st[1] ; j<st[1]+d[1] ; j++)
                {
                    for (k=st[2] ; k<st[2]+d[2] ; k++)
                    {
                        fprintf(stderr, "array[%d][%d][%d].c = %d"
                                        " array[%d][%d][%d].l = %d"
                                        " array[%d][%d][%d].d = %e\n",
                            i, j, k, array[i][j][k].c,
                            i, j, k, array[i][j][k].l,
                            i, j, k, array[i][j][k].d);
                    }
                }
            }
            fprintf(stderr, "err_code = %d\n", err_code);
            free(free_ptr);
        }
    }

    /*
     * TEST 8
     *    3 dimensional array of struct.
     */
    {
        int err_code = 0;
        int asize = 0;
        char *free_ptr;
        int d[3] = {1, 2, 3};
        int st[3] = {-1, -1, -1};
        struct s {
            double d;
            int l;
            char c;
        } ***array, s_init = {1, 10, 0};

        fprintf(stderr, "\nTEST 8\n");

        asize = das(sizeof(struct s), 3, d, &err_code);

        if (test == DAA)
        {
            free_ptr = (char *)malloc(asize);

            array = (struct s ***) daa(sizeof(struct s), 3, d, st,
                &err_code, free_ptr, (char *)&s_init);
        }
        else
        {
            array = (struct s ***) daam(sizeof(struct s), 3, d, st,
                &err_code, &free_ptr, (char *)&s_init);
        }

        if (array == NULL)
        {
            fprintf(stderr, "daa(m): error on dynamic allocation. %s\n"
                , daa_errs[err_code]);
        }
        else
        {
            array[-1][0][1].l = 15;
            array[-1][0][1].d = 7.5;
            array[-1][0][1].c = 2;

            fprintf(stderr, "sizeof(struct s) = %d\n", sizeof(struct s));
            fprintf(stderr, "array size = %d\n", asize);
            fprintf(stderr, "array[-1][0][0].l = %d"
                            " array[-1][0][0].d = %e"
                            " array[-1][0][0].c = %d\n",
               array[-1][0][0].l, array[-1][0][0].d, array[-1][0][0].c);
            fprintf(stderr, "array[-1][0][1].l = %d"
                            " array[-1][0][1].d = %e"
                            " array[-1][0][1].c = %d\n",
               array[-1][0][1].l, array[-1][0][1].d, array[-1][0][1].c);
            fprintf(stderr, "err_code = %d\n", err_code);
            free(free_ptr);
        }
    }

    /*
     * TEST 9
     *    1 dimensional int array.
     */
    {
        int i;
        int err_code = 0;
        int asize = 0;
        char *free_ptr;
        int d[1] = {10};
        int st[1] = {1};
        int init = 12;
        int *array;

        fprintf(stderr, "\nTEST 9\n");

        asize = das(sizeof(int), 1, d, &err_code);

        if (test == DAA)
        {
            free_ptr = (char *)malloc(asize);

            array = (int *) daa(sizeof(int), 1, d, st,
                &err_code, free_ptr, (char *)&init);
        }
        else
        {
            array = (int *) daam(sizeof(int), 1, d, st,
                &err_code, &free_ptr, (char *)&init);
        }

        if (array == NULL)
        {
            fprintf(stderr, "daa(m): error on dynamic allocation. %s\n"
                , daa_errs[err_code]);
        }
        else
        {
            array[10] = 4;

            fprintf(stderr, "sizeof(int) = %d\n", sizeof(int));
            fprintf(stderr, "array size = %d\n", asize);
            for (i=st[0]; i<st[0]+d[0] ; i++)
            {
                fprintf(stderr, "array[%2d] = %d\n", i, array[i]);
            }
            fprintf(stderr, "err_code = %d\n", err_code);
            free(free_ptr);
        }
    }

    /*
     * TEST 10
     *    2 dimensional int array.
     */
    {
        int i, j;
        int err_code = 0;
        int asize = 0;
        char *free_ptr;
        int d[2] = {2, 2};
        int st[2] = {-20, 20};
        int data;
        int init = 0;
        int **array;

        fprintf(stderr, "\nTEST 10\n");

        asize = das(sizeof(int), 2, d, &err_code);

        if (test == DAA)
        {
            free_ptr = (char *)malloc(asize);

            array = (int **) daa(sizeof(int), 2, d, st,
                &err_code, free_ptr, (char *)&init);
        }
        else
        {
            array = (int **) daam(sizeof(int), 2, d, st,
                &err_code, &free_ptr, (char *)&init);
        }

        if (array == NULL)
        {
            fprintf(stderr, "daa(m): error on dynamic allocation. %s\n"
                , daa_errs[err_code]);
        }
        else
        {
            data = 1;
            for (i=st[0] ; i<d[0]+st[0] ; i++)
            {
                for (j=st[1] ; j<d[1]+st[1] ; j++)
                {
                    array[i][j] = data++;
                }    
            }

            fprintf(stderr, "sizeof(int) = %d\n", sizeof(int));
            fprintf(stderr, "array size = %d\n", asize);
            for (i=st[0] ; i<d[0]+st[0] ; i++)
            {
                for (j=st[1] ; j<d[1]+st[1] ; j++)
                {
                    fprintf(stderr, "array[%3d][%3d] = %d\n"
                        , i, j, array[i][j]);
                }    
            }
            fprintf(stderr, "err_code = %d\n", err_code);
            free(free_ptr);
        }
    }

    /*
     * TEST 11
     *    3 dimensional int array.
     */
    {
        int i, j, k;
        int err_code = 0;
        int asize = 0;
        char *free_ptr;
        int d[3] = {3, 3, 3};
        int st[3] = {-10, -20, -30};
        int data;
        int init = 2;
        int ***array;

        fprintf(stderr, "\nTEST 11\n");

        asize = das(sizeof(int), 3, d, &err_code);

        if (test == DAA)
        {
            free_ptr = (char *)malloc(asize);

            array = (int ***) daa(sizeof(int), 3, d, st,
                &err_code, free_ptr, (char *)&init);
        }
        else
        {
            array = (int ***) daam(sizeof(int), 3, d, st,
                &err_code, &free_ptr, (char *)&init);
        }

        if (array == NULL)
        {
            fprintf(stderr, "daa(m): error on dynamic allocation. %s\n"
                , daa_errs[err_code]);
        }
        else
        {
            data = 0;
            for (i=st[0] ; i<st[0]+d[0] ; i++)
            {
                for (j=st[1] ; j<st[1]+d[1] ; j++)
                {
                    for (k=st[2] ; k<st[2]+d[2] ; k++)
                    {
                        array[i][j][k] = data++;
                    }
                }    
            }

            fprintf(stderr, "sizeof(int) = %d\n", sizeof(int));
            fprintf(stderr, "array size = %d\n", asize);
            for (i=st[0] ; i<st[0]+d[0] ; i++)
            {
                for (j=st[1] ; j<st[1]+d[1] ; j++)
                {
                    for (k=st[2] ; k<st[2]+d[2] ; k++)
                    {
                        fprintf(stderr, "array[%3d][%3d][%3d] = %d\n"
                            , i, j, k, array[i][j][k]);
                    }
                }    
            }
            fprintf(stderr, "err_code = %d\n", err_code);
            free(free_ptr);
        }
    }

    /*
     * TEST 12
     *    3 dimensional union array.
     */
    {
        int err_code = 0;
        int asize = 0;
        char *free_ptr;
        int d[3] = {2, 5, 6};
        int st[3] = {0, 0, 0};
        union s {
            int l;
            double d;
        } ***array, s_init = {5};

        fprintf(stderr, "\nTEST 12\n");

        asize = das(sizeof(union s), 3, d, &err_code);

        if (test == DAA)
        {
            free_ptr = (char *)malloc(asize);

            array = (union s ***) daa(sizeof(union s), 3, d, st,
                &err_code, free_ptr, (char *)&s_init);
        }
        else
        {
            array = (union s ***) daam(sizeof(union s), 3, d, st,
                &err_code, &free_ptr, (char *)&s_init);
        }

        if (array == NULL)
        {
            fprintf(stderr, "daa(m): error on dynamic allocation. %s\n"
                , daa_errs[err_code]);
        }
        else
        {
            fprintf(stderr, "sizeof(union s) = %d\n", sizeof(union s));
            fprintf(stderr, "array size = %d\n", asize);

            fprintf(stderr, "array[1][4][4].l = %d\n", array[1][4][4].l);

            array[1][4][5].l = 7;
            fprintf(stderr, "array[1][4][5].l = %d\n", array[1][4][5].l);

            array[1][4][5].d = 7.5;
            fprintf(stderr, "array[1][4][5].d = %e\n", array[1][4][5].d);
            fprintf(stderr, "err_code = %d\n", err_code);
            free(free_ptr);
        }
    }

    /*
     * TEST 13
     *    3 dimensional enum array.
     */
    {
        int err_code = 0;
        int asize = 0;
        char *free_ptr;
        int d[3] = {2, 5, 6};
        int st[3] = {0, 0, 0};
        enum e {a,b,c} ***array, e_init = {c};

        fprintf(stderr, "\nTEST 13\n");

        asize = das(sizeof(enum e), 3, d, &err_code);

        if (test == DAA)
        {
            free_ptr = (char *)malloc(asize);

            array = (enum e ***) daa(sizeof(enum e), 3, d, st,
                &err_code, free_ptr, (char *)&e_init);
        }
        else
        {
            array = (enum e ***) daam(sizeof(enum e), 3, d, st,
                &err_code, &free_ptr, (char *)&e_init);
        }

        if (array == NULL)
        {
            fprintf(stderr, "daa(m): error on dynamic allocation. %s\n"
                , daa_errs[err_code]);
        }
        else
        {
            fprintf(stderr, "sizeof(enum e) = %d\n",sizeof(enum e));
            fprintf(stderr, "array size = %d\n", asize);

            fprintf(stderr, "array[1][4][4] = %d\n", array[1][4][4]);

            array[1][4][5] = a;
            fprintf(stderr, "array[1][4][5] = %d\n", array[1][4][5]);

            array[1][4][5] = b;
            fprintf(stderr, "array[1][4][5] = %d\n", array[1][4][5]);
            fprintf(stderr, "err_code = %d\n", err_code);
            free(free_ptr);
        }
    }

    /*
     * TEST 14
     *    2 dimensional short array.
     */
    {
        int i, j;
        int err_code = 0;
        int asize = 0;
        char *free_ptr;
        int d[2] = {15, 5};
        int st[2] = {-1, -1};
        short init = 11;
        short **array;

        fprintf(stderr, "\nTEST 14\n");

        asize = das(sizeof(short), 2, d, &err_code);

        if (test == DAA)
        {
            free_ptr = (char *)malloc(asize);

            array = (short **) daa(sizeof(short), 2, d, st,
                &err_code, free_ptr, (char *)&init);
        }
        else
        {
            array = (short **) daam(sizeof(short), 2, d, st,
                &err_code, &free_ptr, (char *)&init);
        }

        if (array == NULL)
        {
            fprintf(stderr, "daa(m): error on dynamic allocation. %s\n"
                , daa_errs[err_code]);
        }
        else
        {
            array[5][3] = 4;

            fprintf(stderr, "sizeof(short) = %d\n", sizeof(short));
            fprintf(stderr, "array size = %d\n", asize);
            for (i=st[0] ; i<st[0]+d[0] ; i++)
            {
                for (j=st[1]; j<st[1]+d[1] ; j++)
                {
                    fprintf(stderr, "array[%2d][%2d] = %d\n"
                        , i, j, array[i][j]);
                }
            }
            fprintf(stderr, "err_code = %d\n", err_code);
            free(free_ptr);
        }
    }

    /*
     * TEST 15
     *    2 dimensional unsigned short array.
     */
    {
        int i, j;
        int err_code = 0;
        int asize = 0;
        char *free_ptr;
        int d[2] = {4, 5};
        int st[2] = {0, 1};
        unsigned short init = 13;
        unsigned short **array;

        fprintf(stderr, "\nTEST 15\n");

        asize = das(sizeof(unsigned short), 2, d, &err_code);

        if (test == DAA)
        {
            free_ptr = (char *)malloc(asize);

            array = (unsigned short **) daa(sizeof(unsigned short), 2, d, st,
                &err_code, free_ptr, (char *)&init);
        }
        else
        {
            array = (unsigned short **) daam(sizeof(unsigned short), 2, d, st,
                &err_code, &free_ptr, (char *)&init);
        }

        if (array == NULL)
        {
            fprintf(stderr, "daa(m): error on dynamic allocation. %s\n"
                , daa_errs[err_code]);
        }
        else
        {
            array[3][5] = 5;

            fprintf(stderr, "sizeof(unsigned short) = %d\n"
                , sizeof(unsigned short));
            fprintf(stderr, "array size = %d\n", asize);
            for (i=st[0] ; i<st[0]+d[0] ; i++)
            {
                for (j=st[1] ; j<st[1]+d[1] ; j++)
                {
                    fprintf(stderr, "array[%2d][%2d] = %d\n"
                        , i, j, array[i][j]);
                }
            }
            fprintf(stderr, "err_code = %d\n", err_code);
            free(free_ptr);
        }
    }

    /*
     * TEST 16
     *    1 dimensional array of 3 four meg char strings.
     */
    {
        int i;
        int err_code = 0;
        int asize = 0;
        char *free_ptr;
        int d[1] = {3};
        int st[1] = {0};

        #define FOUR_MEGS (4*1024*1024)

	/* four megs of unsigned char */
	typedef struct
	{
	    unsigned char s[FOUR_MEGS];
	} STRING_FOUR_MEGS;

        STRING_FOUR_MEGS *array;



        fprintf(stderr, "\nTEST 16\n");

        asize = das(sizeof(STRING_FOUR_MEGS), 1, d, &err_code);

        if (test == DAA)
        {
            free_ptr = (char *)malloc(asize);

	    array = (STRING_FOUR_MEGS *) daa(sizeof(STRING_FOUR_MEGS), 1, d, st,
	        &err_code, free_ptr, (char *)NULL);
        }
        else
        {
	    array = (STRING_FOUR_MEGS *) daam(sizeof(STRING_FOUR_MEGS), 1, d, st,
	        &err_code, &free_ptr, (char *)NULL);
        }

        if (array == NULL)
        {
            fprintf(stderr, "daa(m): error on dynamic allocation. %s\n"
                , daa_errs[err_code]);
        }
        else
        {
            memset((unsigned char *)&array[0], 0, FOUR_MEGS);
            memset((unsigned char *)&array[1], 1, FOUR_MEGS);
            memset((unsigned char *)&array[2], 2, FOUR_MEGS);

            strcpy((char *)&array[0], "AXXXXXXXXX");
            strcpy((char *)&array[1], "XXXXBXXXXX");
            strcpy((char *)&array[2], "XXXXXXXXXC");

            fprintf(stderr, "sizeof(STRING_FOUR_MEGS) = %d\n"
                , sizeof(STRING_FOUR_MEGS));
            fprintf(stderr, "array size = %d\n", asize);
            for (i=st[0]; i<st[0]+d[0] ; i++)
            {
                fprintf(stderr, "array[%2d] = %s\n", i, array[i]);
            }
            fprintf(stderr, "err_code = %d\n", err_code);
            free(free_ptr);
        }
    }
}

#endif /* DAA_TEST */

